#ifndef SCOPEDHANDLE_H_
#define SCOPEDHANDLE_H_

// Class that handles automatic deletion of handle values
template<typename T, BOOL (__stdcall *DestroyFunc)(T), T defaultValue>
class GenericScopedHandle
{
public:
	// Constructors

	GenericScopedHandle() : _handle(defaultValue)
	{
	}

	GenericScopedHandle(T handle) : _handle(handle)
	{
	}

	// Close the handle when the object is deleted
	virtual ~GenericScopedHandle()
	{
		if( _handle != defaultValue )
		{
			OutputDebugString(TEXT("Closing handle\n"));
			DestroyFunc(_handle);
		}
	}

	// Assignment operator
	GenericScopedHandle<T, DestroyFunc, defaultValue>& operator=(T handle)
	{
		if( _handle != defaultValue )
		{
			OutputDebugString(TEXT("Closing handle\n"));
			DestroyFunc(_handle);
		}

		_handle = handle;
		return *this;
	}

	operator T() const
	{
		return _handle;
	}
private:
	// Prevent assignment from another scopedhandle
	GenericScopedHandle<T, DestroyFunc, defaultValue>& operator=(const GenericScopedHandle<T, DestroyFunc, defaultValue> &right)
	{
		return *this;
	}

	// Prevent construction from another scopedhandle
	GenericScopedHandle(const GenericScopedHandle<T, DestroyFunc, defaultValue> &right)
	{
	}

	T _handle;
};

// ScopedHandles for common handle types and their appropriate deletion functions
typedef GenericScopedHandle<HANDLE, CloseHandle, INVALID_HANDLE_VALUE> ScopedHandle;
typedef GenericScopedHandle<HWND, DestroyWindow, NULL> ScopedHWND;
typedef GenericScopedHandle<HICON, DestroyIcon, NULL> ScopedHICON;
typedef GenericScopedHandle<HMENU, DestroyMenu, NULL> ScopedHMENU;

#endif