#include "stdafx.h"
#include "resource.h"
#include "WindowsException.h"
#include "ScopedHandle.h"
#include "NotifyIcon.h"

// Constants
const unsigned int WM_USER_NOTIFYICON = WM_USER + 1;
LPCWSTR g_topLevelClassName = L"BrightnessTopLevelClass";

// Global variables
BYTE g_supportedLevels[256];
DWORD g_supportedLevelCount = 0;
UINT g_taskbarCreatedMessage = 0;
HMENU g_contextMenu = NULL;
NotifyIcon *g_notify = NULL;
UCHAR g_displayPolicy = 0;

// Opens the LCD device
HANDLE GetLcdDevice()
{
	HANDLE lcd = CreateFile(L"\\\\.\\LCD", FILE_ANY_ACCESS, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if( lcd == INVALID_HANDLE_VALUE )
		throw WindowsException(L"An error occurred opening the LCD device: ", GetLastError());

	return lcd;
}

// Queries the LCD device for the supported brightness levels, places it in the global variables g_supportedLevels.
// Returns true if there is at least one supported level, false if not.
bool QuerySupportedLevels()
{
	ZeroMemory(g_supportedLevels, sizeof(g_supportedLevels));
	ScopedHandle lcd = GetLcdDevice();
	if( DeviceIoControl(lcd, IOCTL_VIDEO_QUERY_SUPPORTED_BRIGHTNESS, NULL, 0, g_supportedLevels, 256, &g_supportedLevelCount, NULL) )
	{
		if( g_supportedLevelCount == 0 )
		{
			MessageBox(NULL, L"Your video card or driver does not support brightness control.", L"Brightness", MB_OK | MB_ICONINFORMATION);
			return false;
		}
		else
		{
			return true;
		}
	}
	else
	{
		throw WindowsException(L"An error occurred querying supported brightness levels: ", GetLastError());
	}
}

// Gets the current brightness level.
DISPLAY_BRIGHTNESS QueryCurrentLevel()
{
	DISPLAY_BRIGHTNESS result;
	
	ZeroMemory(&result, sizeof(result));

	ScopedHandle lcd = GetLcdDevice();
	DWORD bytesReceived;
	if( DeviceIoControl(lcd, IOCTL_VIDEO_QUERY_DISPLAY_BRIGHTNESS, NULL, 0, &result, sizeof(result), &bytesReceived, NULL) )
	{
		return result;
	}
	else
	{
		throw WindowsException(L"Could not query current brightness level: ", GetLastError());
	}
}

// Sets the brightness level
void SetCurrentLevel(UCHAR newLevel)
{
	DISPLAY_BRIGHTNESS brightness;
	ZeroMemory(&brightness, sizeof(brightness));
	brightness.ucDisplayPolicy = g_displayPolicy;
	if( (g_displayPolicy & DISPLAYPOLICY_AC) != 0 )
		brightness.ucACBrightness = newLevel;
	if( (g_displayPolicy & DISPLAYPOLICY_DC) != 0 )
		brightness.ucDCBrightness = newLevel;

	ScopedHandle lcd = GetLcdDevice();
	DWORD bytesReturned;
	if( !DeviceIoControl(lcd, IOCTL_VIDEO_SET_DISPLAY_BRIGHTNESS, &brightness, sizeof(brightness), NULL, 0, &bytesReturned, NULL) )
	{
		throw WindowsException(L"Could not set current brightness level: ", GetLastError());
	}
}

// Updates the slider position, as well as the g_displayPolicy variable, based on the current brightness level
void UpdateSliderPos(HWND hwndDlg)
{
	try
	{
		// Get level
		DISPLAY_BRIGHTNESS brightness = QueryCurrentLevel();
		UCHAR value = 0;
		g_displayPolicy = brightness.ucDisplayPolicy;
		if( (g_displayPolicy & DISPLAYPOLICY_AC) != 0 )
			value = brightness.ucACBrightness;
		else
			value = brightness.ucDCBrightness;

		// Find the level in the supported values
		int pos = 0;
		for( DWORD x = 0; x < g_supportedLevelCount; ++x )
		{
			if( g_supportedLevels[x] == value )
			{
				pos = x;
				break;
			}
		}

		// Update slider pos, reversing the position to get a top-to-bottom slider
		SendDlgItemMessage(hwndDlg, IDC_SLIDER, TBM_SETPOS, TRUE, g_supportedLevelCount - pos - 1);
	}
	catch( const wruntime_error &ex )
	{
		MessageBox(NULL, ex.ErrorMessage().c_str(), L"Brightness", MB_OK | MB_ICONERROR);
		PostQuitMessage(EXIT_FAILURE);
		return;
	}
}

// Shows the brightness dialog at the current cursor position
void ShowBrightnessDialog(HWND hwndDlg)
{
	// Get the dialog size
	RECT r;
	GetWindowRect(hwndDlg, &r);
	int width = r.right - r.left;
	int height = r.bottom - r.top;
	// Get the cursor position
	POINT pos;
	GetCursorPos(&pos);
	// Update the slider with the current brightness level
	UpdateSliderPos(hwndDlg);
	// Show the window as a topmost window
	// TODO: Deal with the situation where the taskbar isn't in the bottom-right part of the screen
	SetWindowPos(hwndDlg, HWND_TOPMOST, pos.x-width, pos.y-height, 0, 0, SWP_NOSIZE);
	ShowWindow(hwndDlg, SW_SHOW);
	SetForegroundWindow(hwndDlg);
}

// Window procedure for the dialog
INT_PTR CALLBACK DialogProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch( uMsg )
	{
	case WM_INITDIALOG: 
		// Set up to handle taskbar created messages
		if( g_taskbarCreatedMessage == 0 )
			g_taskbarCreatedMessage = RegisterWindowMessage(L"TaskbarCreated");
		return TRUE;
	case WM_USER_NOTIFYICON: // Mouse event for the notify icon
		if( lParam == WM_LBUTTONUP )
		{
			// Show the dialog on left click
			ShowBrightnessDialog(hwndDlg);
			return TRUE;
		}
		else if( lParam == WM_RBUTTONUP )
		{
			// Show the context menu on right click
			SetForegroundWindow(hwndDlg);
			POINT pos;
			GetCursorPos(&pos);
			TrackPopupMenu(g_contextMenu, 0, pos.x, pos.y, 0, hwndDlg, NULL);
			return TRUE;
		}
		break;
	case WM_ACTIVATE:
		// Hide the window when it loses focus
		if( wParam == WA_INACTIVE )
		{
			ShowWindow(hwndDlg, SW_HIDE);
		}
		return TRUE;
	case WM_COMMAND:
		// Exit the app when the Exit menu is clicked
		if( HIWORD(wParam) == 0 && LOWORD(wParam) == ID_CONTEXT_EXIT )
		{
			PostQuitMessage(EXIT_SUCCESS);
			return TRUE;
		}
		break;
	case WM_VSCROLL: // The slider position changed
		if( LOWORD(wParam) == TB_THUMBTRACK || LOWORD(wParam) == TB_ENDTRACK )
		{
			// HIWORD for the position would only work with THUMBTRACK not ENDTRACK so don't use it
			LRESULT pos = SendDlgItemMessage(hwndDlg, IDC_SLIDER, TBM_GETPOS, 0, 0);
			UCHAR newLevel = g_supportedLevels[g_supportedLevelCount - pos - 1];
			try
			{
				// Update brightness level
				SetCurrentLevel(newLevel);
			}
			catch( const wruntime_error &ex )
			{
				MessageBox(NULL, ex.ErrorMessage().c_str(), L"Brightness", MB_OK | MB_ICONERROR);
				PostQuitMessage(EXIT_FAILURE);
			}
		}
		break;
	default:
		// Check for the taskbar created message.
		if( uMsg == g_taskbarCreatedMessage && g_notify != NULL )
		{
			try
			{
				// Recreate the notify icon
				g_notify->RecreateIcon();
			}
			catch( const wruntime_error &ex )
			{
				MessageBox(NULL, ex.ErrorMessage().c_str(), L"Brightness", MB_OK | MB_ICONERROR);
				PostQuitMessage(EXIT_FAILURE);
			}
			return TRUE;
		}
	}
	return FALSE;
}

// Window procedure for the top-level window. Doesn't do much.
LRESULT CALLBACK WindowProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	return DefWindowProc(hwnd, uMsg, wParam, lParam);
}

// Creates the top-level window, but doesn't show it
HWND CreateTopLevelWindow(HINSTANCE instance)
{
	// Register window class for the top-level window.
	WNDCLASS wc;
	ZeroMemory(&wc, sizeof(wc));
	wc.lpfnWndProc = WindowProc;
	wc.lpszClassName = g_topLevelClassName;
	wc.hInstance = instance;
	if( RegisterClass(&wc) == 0 )
	{
		throw WindowsException(L"Could not register window class: ", GetLastError());
	}

	// Create the window
	HWND hwnd = CreateWindow(g_topLevelClassName, L"BrightnessTopLevelWindow", WS_POPUPWINDOW, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, NULL, NULL, instance, NULL);
	if( hwnd == INVALID_HANDLE_VALUE )
	{
		UnregisterClass(g_topLevelClassName, instance);
		throw WindowsException(L"Could not create window: ", GetLastError());
	}

	return hwnd;
}

// Entry-point function
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	int retCode = 0;

	try
	{
		// Continue only if the driver supports brightness control
		if( QuerySupportedLevels() )
		{
			// An extra, invisible window is used to prevent the dialog from having a taskbar button.
			// This is created here.
			ScopedHWND topLevel = CreateTopLevelWindow(hInstance);
			// Create the dialog, but don't show it.
			ScopedHWND dialog = CreateDialog(hInstance, MAKEINTRESOURCE(IDD_DIALOG1), topLevel, DialogProc);
			if( dialog == NULL )
				throw WindowsException(L"Could not create dialog: ", GetLastError());

			// Set the min/max range of the slider based on the supported levels.
			SendDlgItemMessage(dialog, IDC_SLIDER, TBM_SETRANGE, TRUE, MAKELONG(0, g_supportedLevelCount-1));

			// Load resources for the notify icon (icon, menu)
			ScopedHICON icon = static_cast<HICON>(LoadImage(hInstance, MAKEINTRESOURCE(IDI_APPICON), IMAGE_ICON, 0, 0, LR_DEFAULTCOLOR));
			if( icon == NULL )
				throw WindowsException(L"Could not load icon: ", GetLastError());
			ScopedHMENU menu = LoadMenu(GetModuleHandle(NULL), MAKEINTRESOURCE(IDR_MENU1));
			if( menu == NULL )
				throw WindowsException(L"Could not load menu: ", GetLastError());
			g_contextMenu = GetSubMenu(menu, 0);

			// Create the notify icon
			NotifyIcon notify(icon, dialog, L"Brightness", 0, WM_USER_NOTIFYICON);
			// Pointer to notify icon is used in DialogProc when the taskbar is recreated.
			g_notify = &notify;

			// Enter message loop
			BOOL bRet;
			MSG msg;
			while( (bRet = GetMessage( &msg, NULL, 0, 0 )) != 0 )
			{ 
				if (bRet == -1)
				{
					throw WindowsException(L"Error calling GetMessage: ", GetLastError());
				}
				else if( !(IsWindow(dialog) && IsDialogMessage(dialog, &msg)) )
				{
					TranslateMessage(&msg); 
					DispatchMessage(&msg); 
				}
			}

			// Return the value in the WM_QUIT message
			retCode = static_cast<int>(msg.wParam);
		}
	}
	catch( const wruntime_error &ex )
	{
		MessageBox(NULL, ex.ErrorMessage().c_str(), L"Brightness", MB_OK | MB_ICONERROR);
		retCode = EXIT_FAILURE;
	}

	// Only piece of cleanup not handled by the resource-acquisition pattern.
	UnregisterClass(g_topLevelClassName, hInstance);

	return retCode;
}